from pathlib import Path
from textwrap import dedent

from homely._virtualenv import PythonManager, RepoVirtualenvConfig


class TestReadVirtualenvConfig:
    def test_no_virtualenv_wanted(self, tmp_path: Path) -> None:
        # no virtualenv config is present
        config1 = tmp_path / 'config1.toml'
        config1.write_text(dedent(
            """
            [project]
            name = "some_project"
            """
        ))

        cfg1 = RepoVirtualenvConfig.from_pyproject_toml(config1)
        assert cfg1.want_virtualenv is False

        # virtualenv is turned OFF
        config2 = tmp_path / 'config1.toml'
        config2.write_text(dedent(
            """
            [tool.homely.virtualenv]
            minimum_python_version = "3.10"
            install_python_with = ["uv", "pyenv", "homebrew"]
            use_virtualenv = false
            virtualenv_path = "<desired_venv_path>"
            """
        ))

        cfg2 = RepoVirtualenvConfig.from_pyproject_toml(config2)
        assert cfg2.want_virtualenv is False

    def test_valid_virtualenv_config(self, tmp_path: Path) -> None:
        # minimal config
        config1 = tmp_path / 'config1.toml'
        config1.write_text(dedent(
            """
            [tool.homely.virtualenv]
            use_virtualenv = true
            minimum_python_version = "3.11"
            """
        ))

        cfg1 = RepoVirtualenvConfig.from_pyproject_toml(config1)
        assert cfg1.want_virtualenv is True
        assert cfg1.venv_path is None
        assert cfg1.min_python_version == (3, 11, 0)
        assert cfg1.install_with == []
        assert cfg1.config_warnings == []

        # config with all fields
        config2 = tmp_path / 'config2.toml'
        config2.write_text(dedent(
            """
            [tool.homely.virtualenv]
            use_virtualenv = true
            minimum_python_version = "3.10"
            install_python_with = ["uv", "pyenv", "homebrew"]
            virtualenv_path = "/path/to/venv"
            """
        ))

        cfg2 = RepoVirtualenvConfig.from_pyproject_toml(config2)
        assert cfg2.want_virtualenv is True
        assert cfg2.venv_path == Path("/path/to/venv")
        assert cfg2.min_python_version == (3, 10, 0)
        assert cfg2.install_with == [
            PythonManager.UV,
            PythonManager.PYENV,
            PythonManager.HOMEBREW,
        ]
        assert cfg2.config_warnings == []

    def test_virtualenv_config_needs_python_version_specifier(self, tmp_path: Path) -> None:
        # config with all fields
        config1 = tmp_path / 'config1.toml'
        config1.write_text(dedent(
            """
            [tool.homely.virtualenv]
            use_virtualenv = true
            install_python_with = ["uv"]
            """
        ))

        cfg1 = RepoVirtualenvConfig.from_pyproject_toml(config1)
        assert cfg1.want_virtualenv is True
        assert any("minimum_python_version is missing" in warning for warning in cfg1.config_warnings)

    def test_virtualenv_config_notices_invalid_installer(self, tmp_path: Path) -> None:
        # invalid installer name
        config1 = tmp_path / 'config1.toml'
        config1.write_text(dedent(
            """
            [tool.homely.virtualenv]
            use_virtualenv = true
            minimum_python_version = "3.10"
            install_python_with = ["magic"]
            """
        ))

        cfg1 = RepoVirtualenvConfig.from_pyproject_toml(config1)
        assert cfg1.want_virtualenv is True
        assert any("install_python_with contains" in warning for warning in cfg1.config_warnings)

        # invalid installer value (string instead of list)
        config2 = tmp_path / 'config2.toml'
        config2.write_text(dedent(
            """
            [tool.homely.virtualenv]
            use_virtualenv = true
            minimum_python_version = "3.10"
            install_python_with = "uv"
            """
        ))

        cfg2 = RepoVirtualenvConfig.from_pyproject_toml(config2)
        assert cfg2.want_virtualenv is True
        assert any("must be a list" in warning for warning in cfg2.config_warnings)



from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional


class PythonManager(Enum):
    UV = "uv"
    PYENV = "pyenv"
    HOMEBREW = "homebrew"


@dataclass
class RepoVirtualenvConfig:
    want_virtualenv: bool
    venv_path: Optional[Path]
    min_python_version: tuple[int, int, int]
    install_with: list[PythonManager]
    config_warnings: list[str]

    @classmethod
    def from_pyproject_toml(cls, pyproject_path: Path) -> "RepoVirtualenvConfig":
        try:
            # we need a type-ignore here because tomllib is only in Python
            # 3.11+ and mypy is configured to check against 3.10 stdlib.
            import tomllib  # type: ignore
        except ImportError:
            import tomli as tomllib

        with pyproject_path.open('rb') as f:
            try:
                toml_dict = tomllib.load(f)
            except tomllib.TOMLDecodeError:
                # FIXME: raise a better exception here
                raise

        try:
            venv_table = toml_dict['tool']['homely']['virtualenv']
        except KeyError:
            return RepoVirtualenvConfig(
                want_virtualenv=False,
                venv_path=None,
                min_python_version=(0, 0, 0),
                install_with=[],
                config_warnings=[],
            )

        config_warnings = []

        try:
            venv_path = Path(venv_table['virtualenv_path'])
        except TypeError:
            config_warnings.append("Error parsing tool.homely.virtualenv.virtualenv_path. It must be a string file path.")
            venv_path = None
        except KeyError:
            venv_path = None

        try:
            min_python_version = _parse_python_version(
                venv_table['minimum_python_version'],
                "tool.homely.virtualenv.minimum_python_version",
                config_warnings,
            )
        except KeyError:
            min_python_version = (0, 0, 0)
            config_warnings.append(
                "tool.homely.virtualenv.minimum_python_version is missing.",
            )

        install_with = []
        install_with_raw = venv_table.get('install_python_with', [])
        if not isinstance(install_with_raw, list):
            config_warnings.append(
                "Error parsing tool.homely.virtualenv.install_python_with."
                " It must be a list of zero or more of these strings: 'uv', 'pyenv', 'homebrew'.")
        else:
            for value in install_with_raw:
                if value == 'uv':
                    install_with.append(PythonManager.UV)
                elif value == 'pyenv':
                    install_with.append(PythonManager.PYENV)
                elif value == 'homebrew':
                    install_with.append(PythonManager.HOMEBREW)
                else:
                    config_warnings.append(
                        f"tool.homely.virtualenv.install_python_with contains invalid value {value!r}."
                        " It must be a list of zero or more of these strings: 'uv', 'pyenv', 'homebrew'."
                    )

        return RepoVirtualenvConfig(
            want_virtualenv=venv_table.get('use_virtualenv', False),
            venv_path=venv_path,
            min_python_version=min_python_version,
            install_with=install_with,
            config_warnings=config_warnings,
        )


def _parse_python_version(version_str: str, label: str, config_warnings: list[str]) -> tuple[int, int, int]:
    if not isinstance(version_str, str):
        config_warnings.append(f"Error parsing {label}. Expected a string like '3.13'.")
        return (0, 0, 0)

    parts = version_str.split('.')

    try:
        major = int(parts[0])
        minor = int(parts[1]) if len(parts) > 1 else 0
        patch = int(parts[2]) if len(parts) > 2 else 0
    except Exception:
        config_warnings.append(f"Error parsing {label}. Expected a string like '3.13'.")
        return (0, 0, 0)

    return (major, minor, patch)



import os
from datetime import datetime, timedelta
from typing import Optional

from homely._engine2 import getengine, getrepoinfo
from homely._ui import entersection, head, note, warn
# allow importing from outside
from homely._utils import haveexecutable  # noqa
from homely._utils import _loadmodule, _repopath2real, _time_interval_to_delta
# TODO: remove these deprecated aliases which I'm still using in my homely
# repos. Note that the cleaners will need some sort of special handling in
# cleanerfromdict() if ever we want to remove these imports
from homely.files import (WHERE_ANY, WHERE_BOT, WHERE_END, WHERE_TOP,  # noqa
                          CleanBlockInFile, CleanLineInFile, WriteFile,
                          blockinfile, download, lineinfile, mkdir, symlink,
                          writefile)


def run(updatehelper):
    getengine().run(updatehelper)


_include_num = 0


def include(pyscript):
    path = _repopath2real(pyscript, getrepoinfo().localrepo)
    if not os.path.exists(path):
        warn("{} not found at {}".format(pyscript, path))
        return

    global _include_num
    _include_num += 1

    name = '__imported_by_homely_{}'.format(_include_num)
    try:
        with entersection("/" + pyscript):
            return _loadmodule(name, path)
    except Exception:
        import traceback
        warn("Error while including {}: {}".format(pyscript,
                                                   traceback.format_exc()))


def section(func=None, quick=False, enabled=True, interval=None):
    delta = None
    if interval:
        delta = _time_interval_to_delta(interval)

    def _decorator(func):
        return _execute_section(
            func,
            is_quick=quick,
            is_enabled=enabled,
            interval=delta,
        )

    if func:
        return _decorator(func)
    else:
        return _decorator


def _execute_section(func, is_quick, is_enabled, interval: Optional[timedelta] = None) -> None:
    name = func.__name__
    engine = getengine()

    if not is_enabled:
        note("Skipping @section {}() due to enabled=False".format(name))
        return

    if engine.quickmode and not is_quick:
        note("Skipping @section {}() due to --quick flag".format(name))
        return

    if interval:
        timeformat = '%Y-%m-%d %H:%M:%S'
        repoinfo = getrepoinfo()
        assert repoinfo is not None
        last_run_fact_name = 'section_last_run:{}:{}'.format(repoinfo.repoid, name)
        lastrun = None
        try:
            lastrun = engine._getfact(last_run_fact_name)
        except KeyError:
            pass
        if lastrun:
            # XXX: type-ignore on this because it doesn't exist in python3.6
            lastruntime: datetime = datetime.strptime(lastrun, timeformat)  # type: ignore

            nextrun = lastruntime + interval
            if datetime.now() < nextrun:
                note("Skipping @section {}(), not due again until {}".format(name, nextrun))
                return

    try:
        with entersection(":" + name + "()"):
            if engine.pushsection(name):
                head("Executing @section {}()".format(name))
                func()
                if interval:
                    engine._setfact(last_run_fact_name, datetime.now().strftime(timeformat))
            else:
                note("Skipping @section {}() due to -o/--only flag".format(name))
    finally:
        engine.popsection(name)



import os.path
import sys
from io import StringIO

import pytest

from homely._test import contents


@pytest.fixture
def captured_stderr():
    from homely._ui import setstreams

    stream = StringIO()

    setstreams(sys.stdout, stream)

    yield stream

    setstreams(sys.stdout, sys.stderr)


def test_homely_general_include_executes_target_scripts(HOME, testrepo):
    from homely._test import run_update_all

    contents(
        testrepo.remotepath + '/HOMELY.py',
        """
        from homely.general import include
        include("test1/HOMELY.py")
        include("test2/HOMELY.py")
        """
    )

    contents(
        testrepo.remotepath + '/test1/HOMELY.py',
        """
        from homely.files import writefile
        with writefile("~/file1.txt") as f: f.write("I.\\n")
        """,
        mkdir=True,
    )

    contents(
        testrepo.remotepath + '/test2/HOMELY.py',
        """
        from homely.files import writefile
        with writefile("~/file2.txt") as f: f.write("II.\\n")
        """,
        mkdir=True,
    )

    run_update_all(
        # pull first so we get the new HOMELY.py script
        pullfirst=True,
    )

    assert contents(HOME + '/file1.txt') == "I.\n"
    assert contents(HOME + '/file2.txt') == "II.\n"


def test_homely_general_include_returns_target_modules(HOME, testrepo):
    from homely._test import run_update_all

    contents(
        testrepo.remotepath + '/HOMELY.py',
        """
        from homely.general import include

        mod1 = include("mod1/HOMELY.py")
        abc = include("a/b/c/HOMELY.py")

        assert mod1.THE_NUMBER == "Five"
        abc.do_something()
        """
    )

    contents(
        testrepo.remotepath + '/mod1/HOMELY.py',
        """
        THE_NUMBER = 'Five'
        """,
        mkdir=True,
    )

    contents(
        testrepo.remotepath + '/a/b/c/HOMELY.py',
        """
        from homely.files import writefile
        def do_something():
            with writefile("~/abc.txt") as f:
                f.write("Now I know my ABC's\\n")
        """,
        mkdir=True,
    )

    run_update_all(
        # pull first so we get the new HOMELY.py script
        pullfirst=True,
    )

    assert contents(HOME + '/abc.txt') == "Now I know my ABC's\n"


def test_homely_general_include_can_import_parent_module(HOME, testrepo):
    from homely._test import run_update_all

    contents(
        testrepo.remotepath + '/HOMELY.py',
        """
        from homely.general import include
        THE_WORD = "bonza"
        include("subdir/HOMELY.py")
        """
    )

    contents(
        testrepo.remotepath + '/subdir/HOMELY.py',
        """
        from homely.files import writefile
        import HOMELY
        with writefile('~/the_word.txt') as f: f.write(HOMELY.THE_WORD)
        """,
        mkdir=True,
    )

    run_update_all(
        # pull first so we get the new HOMELY.py script
        pullfirst=True,
    )

    assert contents(HOME + '/the_word.txt') == "bonza"


def test_homely_general_include_is_always_relative_to_repo_root(HOME, testrepo):
    """
    This test also ensures the main "HOMELY" import is always the root module
    """
    from homely._test import run_update_all

    contents(
        testrepo.remotepath + '/HOMELY.py',
        """
        from homely.general import include
        THE_WORD = 'flamingo'
        include("a/HOMELY.py")
        """
    )

    contents(
        testrepo.remotepath + '/a/HOMELY.py',
        """
        THE_WORD = 'flamingo'
        from homely.general import include
        include("a/b/HOMELY.py")
        """,
        mkdir=True,
    )

    contents(
        testrepo.remotepath + '/a/b/HOMELY.py',
        """
        from homely.general import include
        include("a/b/c/HOMELY.py")
        """,
        mkdir=True,
    )

    contents(
        testrepo.remotepath + '/a/b/c/HOMELY.py',
        """
        from homely.files import writefile
        import HOMELY
        with writefile('~/the_word.txt') as f: f.write(HOMELY.THE_WORD)
        """,
        mkdir=True,
    )

    run_update_all(
        # pull first so we get the new HOMELY.py script
        pullfirst=True,
    )

    assert contents(HOME + '/the_word.txt') == "flamingo"


def test_homely_general_include_works_with_multiple_repos(HOME, testrepo, testrepo2):
    """
    Tests three facets of include() when there are multiple dotfiles repos:
    - the script from the correct dotfiles repo is executed
    - the "HOMELY" import inside the included script is the root module for that repo
    - the module returned by include() is the correct module
    """
    from homely._test import run_update_all

    contents(
        testrepo.remotepath + '/HOMELY.py',
        """
        from homely.general import include
        THE_TEXT = 'The Phantom Mentace'
        a = include("a/HOMELY.py")
        assert a.THE_NUMBER == "One"
        """
    )

    contents(
        testrepo.remotepath + '/a/HOMELY.py',
        """
        from homely.files import writefile
        import HOMELY
        with writefile('~/text1.txt') as f: f.write(HOMELY.THE_TEXT)
        THE_NUMBER = "One"
        """,
        mkdir=True,
    )

    contents(
        testrepo2.remotepath + '/HOMELY.py',
        """
        from homely.general import include
        THE_TEXT = 'The Clone Wars'
        a = include("a/HOMELY.py")
        assert a.THE_NUMBER == "Two"
        """
    )

    contents(
        testrepo2.remotepath + '/a/HOMELY.py',
        """
        from homely.files import writefile
        import HOMELY
        with writefile('~/text2.txt') as f: f.write(HOMELY.THE_TEXT)
        THE_NUMBER = "Two"
        """,
        mkdir=True,
    )

    run_update_all(
        # pull first so we get the new HOMELY.py script
        pullfirst=True,
    )

    assert contents(HOME + '/text1.txt') == 'The Phantom Mentace'
    assert contents(HOME + '/text2.txt') == 'The Clone Wars'


def test_homely_general_include_relative_includes_do_not_work(HOME, testrepo, captured_stderr):
    """
    Tests three facets of include() when there are multiple dotfiles repos:
    - the script from the correct dotfiles repo is executed
    - the "HOMELY" import inside the included script is the root module for that repo
    - the module returned by include() is the correct module
    """
    from homely._test import run_update_all

    contents(
        testrepo.remotepath + '/HOMELY.py',
        """
        from homely.general import include
        include("a/HOMELY.py")
        """
    )

    contents(
        testrepo.remotepath + '/a/HOMELY.py',
        """
        from homely.general import include
        from homely.files import writefile
        try:
            include("b/HOMELY.py")
        except ImportError:
            with writefile('~/failed.txt') as f: f.write('the include failed')
        """,
        mkdir=True,
    )

    contents(
        testrepo.remotepath + '/a/b/HOMELY.py',
        """
        from homely.files import writefile
        with writefile('~/file1.txt') as f: f.write("Hello world\\n")
        """,
        mkdir=True,
    )

    # the include() will fail and issue a warn(), so run_update_all() will raise an exception
    with pytest.raises(AssertionError, match=r"run_update\(\) encountered errors or warnings"):
        run_update_all(
            # pull first so we get the new HOMELY.py script
            pullfirst=True,
        )

    # file1.txt is not created because the include did not work
    assert not os.path.exists(HOME + '/file1.txt')

    # the 'failed.txt' is not created because include() does not raise an exception
    assert not os.path.exists(HOME + '/failed.txt')

    # there is an ERR in stderr
    captured_stderr.seek(0)
    stderr = captured_stderr.read()
    assert '] ERR   b/HOMELY.py not found at ' in stderr



# NOTE: this file is python3-only because of asyncio
import asyncio


def _runasync(stdoutfilter, stderrfilter, cmd, **kwargs):
    assert asyncio is not None

    async def _runandfilter(loop, cmd, **kwargs):
        def factory():
            return FilteringProtocol(asyncio.streams._DEFAULT_LIMIT, loop)

        class FilteringProtocol(asyncio.subprocess.SubprocessStreamProtocol):
            _stdout = b""
            _stderr = b""

            def pipe_data_received(self, fd, data):
                if fd == 1:
                    if stdoutfilter:
                        self._stdout = stdoutfilter(self._stdout + data, False)
                    else:
                        self.stdout.feed_data(data)
                elif fd == 2:
                    if stderrfilter:
                        self._stderr = stderrfilter(self._stderr + data, False)
                    else:
                        self.stderr.feed_data(data)
                else:
                    raise Exception("Unexpected fd %r" % fd)

            def pipe_connection_lost(self, fd, exc):
                if fd == 1:
                    if stdoutfilter and self._stdout:
                        stdoutfilter(self._stdout, True)
                elif fd == 2:
                    if stderrfilter and self._stderr:
                        stderrfilter(self._stderr, True)
                return super().pipe_connection_lost(fd, exc)

        transport, protocol = await loop.subprocess_exec(factory,
                                                         *cmd,
                                                         **kwargs)
        process = asyncio.subprocess.Process(transport, protocol, loop)

        # now wait for the process to complete
        out, err = await process.communicate()
        return process.returncode, out, err

    _exception = None

    def handleexception(loop, context):
        nonlocal _exception
        if _exception is None:
            _exception = context["exception"]

    # FIXME: probably shouldn't be using the main loop here
    loop = asyncio.get_event_loop()
    loop.set_exception_handler(handleexception)
    result = loop.run_until_complete(_runandfilter(loop, cmd, **kwargs))
    if _exception:
        raise _exception
    return result



import os
import shutil
import tempfile
import time

from homely._utils import opentext


def withtmpdir(func):
    """Decorator for <func> that generates an extra kwarg tmpdir="/path" which
    is a randomly-generated temp dir. A new temp dir is generated on each call
    to <func>, and the temp dirs are automatically cleaned up when <func> ends.

    Note: this decorator needs to go inside any contextlib.contextmanager
    decorator
    """
    def stdwrapper(*args, **kwargs):
        tmpdir = None
        try:
            tmpdir = tempfile.mkdtemp()
            return func(*args, tmpdir=tmpdir, **kwargs)
        finally:
            if tmpdir and os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)

    def genwrapper(*args, **kwargs):
        tmpdir = None
        try:
            tmpdir = tempfile.mkdtemp()
            for item in func(*args, tmpdir=tmpdir, **kwargs):
                yield item
        finally:
            if tmpdir and os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)
    import inspect
    wrapper = genwrapper if inspect.isgeneratorfunction(func) else stdwrapper
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper


def mkdir_recursive(path_to_check):
    if os.path.exists(path_to_check):
        assert os.path.isdir(path_to_check)
        return

    # check the parent exists first
    mkdir_recursive(os.path.dirname(path_to_check))

    # now mkdir the target folder
    os.mkdir(path_to_check)


def contents(path, new_content=None, strip=True, mkdir=False):
    if new_content is not None:
        if mkdir:
            mkdir_recursive(os.path.dirname(os.path.realpath(path)))

        # if new_content was a triple-quoted python string, try and strip off
        # the indent
        if strip and new_content.startswith('\n'):
            indent = len(new_content) - len(new_content[1:].lstrip(' ')) - 1
            stripped = []
            for line in new_content[1:].split('\n'):
                assert line.startswith(' ' * indent) or line == ''
                stripped.append(line[indent:])
            assert stripped[-1] == ''
            new_content = '\n'.join(stripped)

        with opentext(path, 'w') as f:
            f.write(new_content)
    assert os.path.exists(path)
    with opentext(path, 'r') as f:
        return f.read()


NEXT_FILE = 1


def gettmpfilepath(tmpdir, suffix=".txt"):
    global NEXT_FILE
    try:
        return os.path.join(tmpdir, "tmpfile-%03d%s" % (NEXT_FILE, suffix))
    finally:
        NEXT_FILE += 1


def waitfor(desc, maxtime=1, interval=0.05):
    """
    wait up to maxtime
    """
    yield
    start = time.time()
    while True:
        time.sleep(interval)
        yield
        if time.time() > (start + maxtime):
            raise Exception("Waited too long for: %s" % desc)


def run_update_all(pullfirst=False, cancleanup=False, quick=False) -> None:
    from homely._ui import run_update
    from homely._utils import RepoListConfig

    cfg = RepoListConfig()
    success = run_update(
        list(cfg.find_all()),
        pullfirst=pullfirst,
        only=None,
        cancleanup=cancleanup,
        quick=quick,
    )

    assert success, "run_update() encountered errors or warnings"



from datetime import timedelta

import pytest
from freezegun import freeze_time

from homely._test import contents


@pytest.mark.parametrize('interval', ['2w', '14d', '336h', timedelta(days=14), timedelta(weeks=2)])
def test_homely_update_section_intervals(HOME, testrepo, interval):
    from homely._test import run_update_all

    contents(
        testrepo.remotepath + '/HOMELY.py',
        """
        import datetime  # needed for when we use raw timedelta instances as the interval
        from datetime import date
        from homely.general import section, writefile

        @section(interval={!r})
        def time_section():
            with writefile('~/last-update.txt') as f:
                f.write(str(date.today()))
        """.format(interval)
    )

    with freeze_time('2022-03-01'):
        run_update_all(
            # pull first so we get the new HOMELY.py script
            pullfirst=True,
        )
        assert contents(HOME + '/last-update.txt') == "2022-03-01"

    # run again one week later - there should be no change to the last-update
    # file
    with freeze_time('2022-03-08'):
        run_update_all()
        assert contents(HOME + '/last-update.txt') == "2022-03-01"

    # run again after the interval has elapsed
    with freeze_time('2022-03-15'):
        run_update_all()
        assert contents(HOME + '/last-update.txt') == "2022-03-15"

    # run again - no change again
    with freeze_time('2022-03-20'):
        run_update_all()
        assert contents(HOME + '/last-update.txt') == "2022-03-15"

    # run again - interval has elapsed so the new date is written
    with freeze_time('2022-03-30'):
        run_update_all()
        assert contents(HOME + '/last-update.txt') == "2022-03-30"



import json
import os
import sys

from homely._test import contents, gettmpfilepath


def test_engine_folder_cleanup(tmpdir):
    from homely._engine2 import Engine
    from homely.files import MakeDir

    # first thing ... test mkdir cleanup
    # a temporary file where the engine can store its config
    cfgpath = gettmpfilepath(tmpdir, '.json')

    d1 = os.path.join(tmpdir, 'dir1')
    d2 = os.path.join(tmpdir, 'dir2')
    d2a = os.path.join(d2, 'sub-a')
    d2a1 = os.path.join(d2a, 'supersub-1')
    d2a2 = os.path.join(d2a, 'supersub-2')

    # create the first directory
    e = Engine(cfgpath)
    e.run(MakeDir(d1))
    e.cleanup(e.RAISE)
    del e

    e = Engine(cfgpath)
    e.run(MakeDir(d1))
    e.run(MakeDir(d2))
    e.run(MakeDir(d2a))
    e.run(MakeDir(d2a1))
    e.run(MakeDir(d2a2))
    e.cleanup(e.RAISE)
    del e

    # all dirs should exist now
    check = [d1, d2, d2a, d2a1, d2a2]
    for d in check:
        assert os.path.exists(d)
    assert all(map(os.path.exists, check))

    # make a new engine and make ONLY d2a1 ... this should
    # be enough to keep everything but d1 and d2a2 alive
    e = Engine(cfgpath)
    e.run(MakeDir(d2a2))
    e.cleanup(e.POSTPONE)
    del e

    # all dirs should exist except for d1 and d2a1
    assert all(map(os.path.exists, [d2, d2a, d2a2]))
    assert not any(map(os.path.exists, [d1, d2a1]))

    # make a new engine, add nothing to it, this should be enough to blast away
    # all directories
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    del e

    # all dirs are gone, EXCEPT FOR our original tmpdir, which our cleaners
    # were smart enough to leave behind
    assert not any(map(os.path.exists, [d1, d2, d2a, d2a1, d2a2]))
    assert os.path.exists(tmpdir)

    # FIXME: add a test to see what happens when we want to clean up a dir but
    # there's already files inside it ... we can add a test for BLASTAWAY mode
    # here I guess


def test_symlink_cleanup_interaction(tmpdir):
    from homely._engine2 import Engine
    from homely._errors import CleanupConflict
    from homely.files import MakeDir, MakeSymlink

    cfgpath = gettmpfilepath(tmpdir)

    d1 = os.path.join(tmpdir, 'dir1')
    l1 = os.path.join(tmpdir, 'link1')
    d1da = os.path.join(d1, 'dir-a')
    l1da = os.path.join(l1, 'dir-a')

    # basic symlink
    e = Engine(cfgpath)
    e.run(MakeDir(d1))
    e.run(MakeSymlink(d1, l1))
    e.cleanup(e.RAISE)
    del e

    # make sure they both exist, and the symlink has the correct target
    assert os.path.isdir(d1)
    assert os.path.islink(l1) and os.readlink(l1) == d1

    # mkdir() inside a symlink's target
    e = Engine(cfgpath)
    e.run(MakeDir(d1))
    e.run(MakeSymlink(d1, l1))
    e.run(MakeDir(l1da))
    e.cleanup(e.POSTPONE)
    del e

    # everything should exist, the symlink should have the correct target
    assert os.path.isdir(d1)
    assert os.path.islink(l1) and os.readlink(l1) == d1
    assert os.path.isdir(d1da) and os.path.isdir(l1da)

    # if we try and get rid of the symlink now, the cleaner will hang around
    # because it can't garbage collect the thing right now
    e = Engine(cfgpath)
    e.run(MakeDir(d1))
    e.run(MakeDir(l1da))
    try:
        e.cleanup(e.RAISE)
    except CleanupConflict as err:
        # conflict should be over the symlink's path
        assert err.conflictpath == l1
        # the path should still be needed by `l1da`
        assert err.pathwanter == l1da
    else:
        raise Exception("Expected a cleanup error!")  # noqa
    del e

    # clean everything up now
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    del e
    assert not any(map(os.path.exists, [d1, d1da, l1]))
    os.unlink(cfgpath)

    # now we try the opposite - symlink to a directory and try to clean up the
    # directory
    e = Engine(cfgpath)
    e.run(MakeDir(d1))
    e.run(MakeSymlink(d1, l1))
    e.cleanup(e.RAISE)
    del e
    assert os.path.isdir(d1)
    assert os.path.islink(l1) and os.readlink(l1) == d1

    # the engine will happily clean up the directory, even though there is a
    # symlink pointing to it ... symlinks don't care whether the target exists
    # or not
    e = Engine(cfgpath)
    e.run(MakeSymlink(d1, l1))
    e.cleanup(e.RAISE)
    del e

    # clean up everything now
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    assert not any(map(os.path.exists, [d1, d1da, l1]))
    os.unlink(cfgpath)


def test_lineinfile_usage(tmpdir):
    from homely._engine2 import Engine
    from homely.files import WHERE_END, WHERE_TOP, LineInFile

    cfgpath = gettmpfilepath(tmpdir, '.json')

    f1 = gettmpfilepath(tmpdir)
    # make sure the following types of input raise exceptions
    bad = [
        "",  # empty line
        "something\n",  # a string containing a newline
        "   \t",  # a line containing nothing but whitespace
    ]
    for b in bad:
        try:
            LineInFile(f1, b)
        except Exception:
            # we got an exception
            pass
        else:
            raise Exception("LineInFile should not accept input %s" % repr(b))

    # make sure LineInFile() doesn't override an existing line
    contents(f1, "AAA\nBBB\n")
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "BBB"))
    assert contents(f1) == "AAA\nBBB\n"

    # make sure LineInFile() puts a new line at the end of the file by default
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "CCC"))
    assert contents(f1) == "AAA\nBBB\nCCC\n"

    # make sure LineInFile() doesn't move a line unnecessarily
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "BBB"))
    assert contents(f1) == "AAA\nBBB\nCCC\n"

    # make sure LineInFile() can move a line to the TOP of the file
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "BBB", WHERE_TOP))
    assert contents(f1) == "BBB\nAAA\nCCC\n"

    # make sure LineInFile() can move a line to the END of the file
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "BBB", WHERE_END))
    assert contents(f1) == "AAA\nCCC\nBBB\n"

    # make sure LineInFile() doesn't blow away empty lines
    # - adding to end of file
    contents(f1, "\n\n", strip=False)
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "AAA"))
    assert contents(f1) == "\n\nAAA\n"
    # - adding to start of file
    contents(f1, "\n\n", strip=False)
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "AAA", WHERE_TOP))
    assert contents(f1) == "AAA\n\n\n"
    # - when restoring the file
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    del e
    assert contents(f1) == "\n\n"
    # - replacing something in the middle
    contents(f1, "\n\nAAA\n\n\n", strip=False)
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "AAA"))
    assert contents(f1) == "\n\nAAA\n\n\n"

    # make sure LineInFile() respects existing line endings
    # NOTE: in python2 we always use Universal Newlines when reading the file,
    # which tricks us into using "\n" when writing the file
    if sys.version_info[0] > 2:
        # - windows
        contents(f1, "AAA\r\nBBB\r\n")
        e = Engine(cfgpath)
        e.run(LineInFile(f1, "CCC"))
        assert contents(f1) == "AAA\r\nBBB\r\nCCC\r\n"
        # - mac
        contents(f1, "AAA\rBBB\r")
        e = Engine(cfgpath)
        e.run(LineInFile(f1, "BBB", WHERE_TOP))
        assert contents(f1) == "BBB\rAAA\r"

    # make sure a file that starts empty is left empty after cleanup
    contents(f1, "")
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "AAA"))
    assert contents(f1) == "AAA\n"
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    assert contents(f1) == ""


def test_lineinfile_cleanup_interaction(tmpdir):
    from homely._engine2 import Engine
    from homely.files import LineInFile, MakeDir

    # a temporary file where the engine can store its config
    cfgpath = gettmpfilepath(tmpdir, '.json')
    f1 = os.path.join(tmpdir, 'f1.txt')
    d1 = os.path.join(tmpdir, 'f1.txt.dir')
    d1f1 = os.path.join(d1, 'f-1.txt')
    d2 = os.path.join(tmpdir, 'dir2')
    d2f1 = os.path.join(d2, 'f-1.txt')
    d3 = os.path.join(tmpdir, 'dir3')
    d3d1 = os.path.join(d3, 'sub-1')
    d3d2 = os.path.join(d3, 'sub-2')
    d3d1f1 = os.path.join(d3d1, 'somefile.txt')
    d3d2f1 = os.path.join(d3d2, 'somefile.txt')

    # make the dir and the file, make sure they both exist
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "AAA"))
    e.run(MakeDir(d1))
    e.cleanup(e.RAISE)
    del e
    assert os.path.isdir(d1)
    assert contents(f1) == "AAA\n"

    # remake the engine without the dir, make sure the dir isn't kept around by
    # the file
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "AAA"))
    e.cleanup(e.POSTPONE)
    del e
    assert not os.path.exists(d1)
    assert contents(f1) == "AAA\n"

    # make a new file in the directory, make sure they take ownership of the
    # directory
    e = Engine(cfgpath)
    e.run(MakeDir(d1))
    e.run(LineInFile(d1f1, "AAA"))
    e.run(LineInFile(d1f1, "BBB"))
    e.cleanup(e.POSTPONE)
    del e
    # the directory should still exist, the old file should not, and the new
    # file should contain both lines
    assert os.path.isdir(d1)
    assert not os.path.exists(f1)
    assert contents(d1f1) == "AAA\nBBB\n"

    # if we do the same thing again but without the MakeDir, we should get
    # exactly the same result
    e = Engine(cfgpath)
    e.run(LineInFile(d1f1, "AAA"))
    e.run(LineInFile(d1f1, "BBB"))
    e.cleanup(e.POSTPONE)
    del e
    # the directory should still exist, the old file should not, and the new
    # file should contain both lines
    assert os.path.isdir(d1)
    assert not os.path.exists(f1)
    assert contents(d1f1) == "AAA\nBBB\n"

    # the important thing to note is that the engine still knows that it needs
    # to clean up the directory later
    e = Engine(cfgpath)
    assert e.pathstoclean()[d1] == e.TYPE_FOLDER_ONLY
    del e

    # if we get rid of one of the LineInFile() items, the file and dir still
    # hang around
    e = Engine(cfgpath)
    e.run(LineInFile(d1f1, "AAA"))
    e.cleanup(e.POSTPONE)
    del e
    assert contents(d1f1) == "AAA\n"

    # now, when we get rid of the last LineInFile() items, everything will be
    # cleaned up
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    del e
    assert not os.path.exists(d1)
    os.unlink(cfgpath)

    # if we make a dir ourselves, a LineInFile() will not clean it up
    assert not os.path.exists(d2)
    os.mkdir(d2)
    e = Engine(cfgpath)
    e.run(LineInFile(d2f1, "AAA"))
    del e
    assert contents(d2f1) == "AAA\n"
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    del e
    assert os.path.isdir(d2) and not os.path.exists(d2f1)
    os.rmdir(d2)

    # if we make a file ourselves, LineInFile() will not try to clean it up
    assert not os.path.exists(f1)
    contents(f1, "")
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "AAA"))
    del e
    assert contents(f1) == "AAA\n"
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    del e
    assert contents(f1) == ""

    # if two LineInFile() items share a parent dir (even the parent dir's
    # parent, and so on), and that parent dir was made by a MakeDir(), then it
    # gets cleaned up eventually when both of the LineInFile() items go away
    # - create the folder structure
    assert not os.path.exists(d3)
    os.unlink(cfgpath)
    e = Engine(cfgpath)
    e.run(MakeDir(d3))
    e.run(MakeDir(d3d1))
    e.run(MakeDir(d3d2))
    e.cleanup(e.RAISE)
    del e
    assert all(map(os.path.isdir, [d3, d3d1, d3d2]))
    assert not any(map(os.path.exists, [d3d1f1, d3d2f1]))
    # - scrap the folder structure, but make files that depend on those folders
    e = Engine(cfgpath)
    e.run(LineInFile(d3d1f1, "AAA"))
    e.run(LineInFile(d3d2f1, "BBB"))
    e.cleanup(e.POSTPONE)
    del e
    assert contents(d3d1f1) == "AAA\n" and contents(d3d2f1) == "BBB\n"
    # - scrap the first file, and check that only things needed for the 2nd
    # file remain
    e = Engine(cfgpath)
    e.run(LineInFile(d3d2f1, "BBB"))
    e.cleanup(e.POSTPONE)
    del e
    assert not os.path.exists(d3d1)
    assert contents(d3d2f1) == "BBB\n"
    # - scrap the 2nd file and make sure everything else disappears
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    del e
    assert not os.path.exists(d3)


def test_blockinfile_lineinfile_cleanup_interaction(tmpdir):
    '''
    Test that when a LineInFile is cleaned up, any other LineInFile() or
    BlockInFile() that affected the same file, needs a chance to re-check
    whether it is still valid. If *any* of the remaining Helpers needs to be
    reapplied, then they *all* need to be reapplied.
    '''
    from homely._engine2 import Engine
    from homely.files import BlockInFile, LineInFile

    cfgpath = gettmpfilepath(tmpdir, '.json')
    f1 = gettmpfilepath(tmpdir, '.txt')

    def bif(filename, lines):
        # shorthand constructor for BlockInFile()
        return BlockInFile(filename, lines, None, "PRE", "POST")

    # check that a LineInFile followed by a BlockInFile that both try to add
    # the same line will result in the file containing both things, even if
    # you rerun it many times
    for i in range(3):
        e = Engine(cfgpath)
        e.run(LineInFile(f1, "AAA"))
        e.run(bif(f1, ["AAA"]))
        e.cleanup(e.RAISE)
        del e
        assert contents(f1) == "AAA\nPRE\nAAA\nPOST\n"

    # check that a BlockInFile followed by a LineInFile that adds the same line
    # will result in the LineInFile not having any effect, even if you rerun it
    # many times
    os.unlink(f1)
    for i in range(3):
        e = Engine(cfgpath)
        e.run(bif(f1, ["AAA"]))
        e.run(LineInFile(f1, "AAA"))
        e.cleanup(e.RAISE)
        del e
        assert contents(f1) == "PRE\nAAA\nPOST\n"

    # check that removing the LineInFile doesn't destroy the contents added by
    # BlockInFile
    e = Engine(cfgpath)
    e.run(bif(f1, ["AAA"]))
    e.cleanup(e.RAISE)
    del e
    assert contents(f1) == "PRE\nAAA\nPOST\n"

    # put both things back in ...
    e = Engine(cfgpath)
    e.run(bif(f1, ["AAA"]))
    e.run(LineInFile(f1, "AAA"))
    e.cleanup(e.RAISE)
    del e
    assert contents(f1) == "PRE\nAAA\nPOST\n"

    # test that the LineInFile keeps itself there after removing the
    # BlockInFile
    e = Engine(cfgpath)
    e.run(LineInFile(f1, "AAA"))
    e.cleanup(e.RAISE)
    del e
    assert contents(f1) == "AAA\n"


def test_cleanup_everything(tmpdir):
    '''
    test that recreating the engine, running nothing on it, then calling
    cleanup() will remove all of things that might be lying around
    '''
    from homely._engine2 import Engine
    from homely.files import LineInFile, MakeDir, MakeSymlink

    cfgpath = gettmpfilepath(tmpdir, '.json')
    d1 = gettmpfilepath(tmpdir, '.d')
    d1f1 = os.path.join(d1, 'sub-file.txt')
    l1 = gettmpfilepath(tmpdir, '.lnk')

    e = Engine(cfgpath)
    e.run(MakeDir(d1))
    e.run(MakeSymlink(d1, l1))
    e.run(LineInFile(d1f1, "AAA"))
    e.cleanup(e.RAISE)
    del e

    assert os.path.isdir(d1)
    assert contents(d1f1, "AAA\n")
    assert os.path.islink(l1) and os.readlink(l1) == d1

    # if we recreate the engine with nothing on it, it should clean everything
    # up
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)

    assert not os.path.exists(d1)
    assert not os.path.islink(l1)


def test_partial_run_cleanup(tmpdir):
    from homely._engine2 import Engine
    from homely.files import MakeDir

    cfgpath = gettmpfilepath(tmpdir, '.json')
    d1 = os.path.join(tmpdir, 'dir1')
    d2 = os.path.join(tmpdir, 'dir2')

    # simulate adding our first repo which just creates d1
    e = Engine(cfgpath)
    e.run(MakeDir(d1))
    e.cleanup(e.RAISE)
    del e
    assert os.path.isdir(d1) and not os.path.isdir(d2)

    # simulate adding a 2nd repo that would just created 2 ... this shouldn't
    # trigger a cleanup though
    e = Engine(cfgpath)
    e.run(MakeDir(d2))
    del e
    assert os.path.isdir(d1) and os.path.isdir(d2)

    # pretend that the first repo is removed, and do a full update
    e = Engine(cfgpath)
    e.run(MakeDir(d2))
    e.cleanup(e.RAISE)
    assert not os.path.isdir(d1) and os.path.isdir(d2)


def test_writefile_usage(tmpdir):
    from homely._engine2 import Engine
    from homely.files import LineInFile
    from homely.general import WriteFile, writefile

    cfgpath = gettmpfilepath(tmpdir, '.json')

    # the file we'll be playing with
    f1 = os.path.join(tmpdir, 'f1.txt')
    f2 = os.path.join(tmpdir, 'f2.txt')
    f3 = os.path.join(tmpdir, 'f3.json')

    # use LineInFile to put stuff in the file
    e = Engine(cfgpath)
    e.run(LineInFile(f1, 'AAA'))
    e.run(LineInFile(f1, 'BBB'))
    del e
    assert contents(f1) == "AAA\nBBB\n"

    # now use WriteFile() to give it new contents
    e = Engine(cfgpath)
    e.run(WriteFile(f1, "BBB\nCCC\n"))
    assert contents(f1) == "BBB\nCCC\n"
    e.cleanup(e.RAISE)
    del e
    # make sure the cleanup didn't blow anything away
    assert contents(f1) == "BBB\nCCC\n"

    # make sure the file is removed on cleanup
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    del e
    assert not os.path.exists(f1)

    contents(f2, "Already here!\n")
    assert os.path.exists(f2)
    e = Engine(cfgpath)
    e.run(WriteFile(f2, "AAA\nBBB\n"))
    e.cleanup(e.RAISE)
    del e
    assert contents(f2) == "AAA\nBBB\n"

    # running a LineInFile() won't clean up what's already there
    e = Engine(cfgpath)
    e.run(LineInFile(f2, "CCC"))
    e.cleanup(e.RAISE)
    del e
    assert contents(f2) == "AAA\nBBB\nCCC\n"

    # note that removing the LineInFile() doesn't clean up the file because we
    # no longer know whether the user put their config in there
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    del e
    assert not os.path.exists(f1)
    assert os.path.exists(f2)

    # now test the context manager
    import homely._engine2
    e = Engine(cfgpath)
    homely._engine2._ENGINE = e
    data = {"z": [3, 4, 5, True], "y": "Hello world", "x": None}
    with writefile(f3) as f:
        if sys.version_info[0] < 3:
            f.write(json.dumps(data, ensure_ascii=False))
        else:
            f.write(json.dumps(data))
    e.cleanup(e.RAISE)
    del e
    assert os.path.exists(f3)
    with open(f3, 'r') as f:
        assert json.loads(f.read()) == data

    # prove that the WriteFile() disappearing results in the file being removed
    e = Engine(cfgpath)
    e.cleanup(e.RAISE)
    del e
    assert not os.path.exists(f3)



import os
import time

from homely._engine2 import Cleaner, Engine, Helper, getengine
from homely._errors import HelperError
from homely._ui import allowinteractive, allowpull, note
from homely._utils import haveexecutable, isnecessarypath
from homely.system import execute


def installpkg(name, wantcmd=None, **methods):
    for key in methods:
        assert key in _METHODS

    # FIXME: make sure the user specifies at least one way to install the thing
    getengine().run(InstallPackage(name, methods, wantcmd))


_ALLOW_INSTALL = True


def setallowinstall(allow_install):
    """
    Configure whether installpkg() InstallPackage() are actually allowed to
    install anything.

    If installing isn't allowed, installpkg() and InstallPackage() will raise
    an error instead of installing the package. This is useful in work
    environment where your local sysadmin wants additional packages managed
    externally by a tool like salt.

    NOTE: This also controls whether InstallFromSource() is allowed to perform
    commands starting with "sudo" - the assumption here is that if
    InstallFromSource() can't run commands as root, it can't install anything.
    Compiling from source and symlinking to ~/bin will still work fine.
    """
    global _ALLOW_INSTALL
    _ALLOW_INSTALL = bool(allow_install)


class InstallFromSource(Helper):
    _title = None
    _source_repo = None
    _clone_to = None
    _real_clone_to = None
    _branch = None
    _tag = None
    _compile = None

    # FIXME: we need a better way to specify whether or not a TTY is needed
    _needs_tty = False

    def __init__(self, source_repo, clone_to):
        super(InstallFromSource, self).__init__()
        self._title = 'Install %s into %s' % (source_repo, clone_to)
        self._source_repo = source_repo
        self._clone_to = clone_to
        self._real_clone_to = os.path.expanduser(clone_to)
        self._symlinks = []

    def select_branch(self, branch_name, expiry=None):
        # possible values of expiry:
        # 0:     always pull and compile
        # -1:    never pull or compile again
        # <int>: pull and compile again after <int> seconds
        if expiry is None:
            expiry = 60 * 60 * 24 * 14
        assert self._tag is None
        assert type(expiry) is int and expiry >= -1
        self._branch = branch_name
        self._expiry = expiry
        self._branchfact = '{}:compile-branch:{}:{}'.format(
            self.__class__.__name__,
            self._real_clone_to,
            branch_name)

    def select_tag(self, tag_name):
        assert self._branch is None
        self._tag = tag_name
        self._expiry = None

    def symlink(self, target, linkname):
        self._symlinks.append((os.path.join(self._real_clone_to, target),
                               os.path.expanduser(linkname)))

    def compile_cmd(self, commands):
        assert self._compile is None
        self._compile = list(commands)
        for cmd in self._compile:
            if cmd[0] == "sudo":
                self._needs_tty = True

    @property
    def description(self):
        return self._title

    def getcleaner(self):
        return

    def affectspath(self, path):
        return isnecessarypath(self._real_clone_to, path)

    def pathsownable(self):
        ret = {self._real_clone_to: Engine.TYPE_FOLDER_ONLY}
        for target, linkname in self._symlinks:
            ret[linkname] = Engine.TYPE_LINK
        return ret

    def getclaims(self):
        return []

    def isdone(self):
        if not os.path.exists(self._real_clone_to):
            return False

        if self._tag:
            # has the correct branch or tag been checked out?
            current = execute(['git', 'tag', '--points-at', 'HEAD'],
                              cwd=self._real_clone_to,
                              stdout=True)[1]
            if self._tag not in map(str, current.splitlines()):
                return False

        # if there's no symlinks, we can't tell if it's done or not
        # TODO: test this before releasing
        if not len(self._symlinks):
            return False

        # do the symlinks exist?
        for target, linkname in self._symlinks:
            if not os.path.islink(linkname):
                return False
            if os.readlink(linkname) != target:
                return False

        # it appears to be done ... yay
        return True

    def makechanges(self):
        assert self._source_repo is not None
        assert self._clone_to is not None
        if not os.path.exists(self._real_clone_to):
            note("Cloning %s" % self._source_repo)
            pull_needed = False
            execute(['git', 'clone', self._source_repo, self._real_clone_to])
        else:
            pull_needed = True
            if not os.path.exists(os.path.join(self._real_clone_to, '.git')):
                raise HelperError("%s is not a git repo" % self._real_clone_to)

        # do we want a particular branch?
        if self._branch:
            execute(['git', 'checkout', self._branch], cwd=self._real_clone_to)
            if pull_needed and allowpull():
                note("Updating %s from %s" %
                     (self._clone_to, self._source_repo))
                execute(['git', 'pull'], cwd=self._real_clone_to)

            # check the branch fact to see if we need to compile again
            factname = self._branchfact
        else:
            assert self._tag is not None
            if pull_needed and allowpull():
                note("Updating %s from %s" %
                     (self._clone_to, self._source_repo))
                # NOTE: we use --force for projects like neovim that have a
                # rolling 'nightly' tag
                execute(['git', 'fetch', '--tags', '--force'],
                        cwd=self._real_clone_to)
            execute(['git', 'checkout', self._tag], cwd=self._real_clone_to)

            # if we used a tag name, create a 'fact' to prevent us re-compiling
            # each time we run
            factname = '{}:compile-tag:{}:{}'.format(
                self.__class__.__name__,
                self._real_clone_to,
                self._tag)

        docompile = False
        if self._compile:
            last_compile, prev_cmds = self._getfact(factname, (0, None))
            what = ("Branch {}".format(self._branch) if self._branch
                    else "Tag {}".format(self._tag))
            if last_compile == 0:
                note("{} has never been compiled".format(what))
                docompile = True
            elif (self._expiry is not None
                  and ((last_compile + self._expiry) < time.time())):
                note("{} is due to be compiled again".format(what))
                docompile = True
            elif prev_cmds != self._compile:
                note("{} needs to be compiled again with new commands"
                     .format(what))
                docompile = True

        # run any compilation commands
        if docompile:
            # FIXME: we probably need to delete all the symlink targets before
            # compiling, as this is our best way of determining that the
            # compilation has failed ...
            stdout = "TTY" if self._needs_tty else None
            for cmd in self._compile:
                if cmd[0] == "sudo" and not _ALLOW_INSTALL:
                    raise HelperError(
                        "%s is not allowed to run commands as root"
                        ", as per setallowinstall()")
                execute(cmd, cwd=self._real_clone_to, stdout=stdout)

            self._setfact(factname, (time.time(), self._compile))

        # create new symlinks
        for source, dest in self._symlinks:
            with note("Ensure symlink exists: %s -> %s" % (source, dest)):
                if os.path.islink(dest):
                    target = os.readlink(dest)
                    if os.path.realpath(target) != os.path.realpath(source):
                        raise HelperError("Symlink %s is not pointing at %s" %
                                          (dest, source))
                    continue
                if os.path.exists(dest):
                    raise HelperError("%s already exists" % dest)
                os.symlink(source, dest)


_METHODS = ('brew', 'yum', 'apt', 'port', 'pacman')
_ASROOT = ('yum', 'port', 'apt', 'pacman')
_INSTALL = {
    'apt': lambda name: ['apt-get', 'install', name, '--quiet',
                         '--assume-yes'],
    'yum': lambda name: ['yum', 'install', name, '--assumeyes'],
    'pacman': lambda name: ['pacman', '-S', '--quiet',
                         '--noconfirm', name],
}
_UNINSTALL = {
    'apt': lambda name: ['apt-get', 'remove', name, '--quiet', '--assume-yes'],
    'yum': lambda name: ['yum', 'erase', name, '--assumeyes'],
    'pacman': lambda name: ['pacman', '-R', '--noconfirm', name],
}


class InstallPackage(Helper):
    def __init__(self, name, methods, wantcmd):
        super(InstallPackage, self).__init__()
        self._name = name
        self._methods = methods
        self._wantcmd = name if wantcmd is None else wantcmd

    def getcleaner(self):
        return PackageCleaner(self._name, self._methods)

    def pathsownable(self):
        return {}

    def isdone(self):
        return haveexecutable(self._wantcmd)

    @property
    def description(self):
        how = [m for m in _METHODS if self._methods.get(m, True)]
        return "Install package %s using %s" % (self._name, how)

    def getclaims(self):
        yield "package:%s" % self._name

    def affectspath(self, path):
        return False

    def makechanges(self):
        # try each method
        for method in _METHODS:
            localname = self._methods.get(method, self._name)

            if localname is False:
                continue

            def getdefaultcmd(name):
                return [method, 'install', name]

            cmd = _INSTALL.get(method, getdefaultcmd)(localname)

            # see if the required executable is installed
            if not haveexecutable(cmd[0]):
                continue

            if not _ALLOW_INSTALL:
                raise HelperError(
                    "InstallPackage() is not allowed to install packages"
                    ", as per setallowinstall()")

            if method in _ASROOT:
                if not allowinteractive():
                    raise HelperError("Need to be able to escalate to root")
                cmd.insert(0, 'sudo')
            execute(cmd)
            # record the fact that we installed this thing ourselves
            factname = 'InstalledPackage:%s:%s' % (method, localname)
            self._setfact(factname, True)
            return
        raise HelperError("No way to install %s" % self._name)


class PackageCleaner(Cleaner):
    def __init__(self, name, methods):
        super(PackageCleaner, self).__init__()
        self._name = name
        self._methods = methods

    def asdict(self):
        return dict(name=self._name, methods=self._methods)

    @classmethod
    def fromdict(class_, data):
        return class_(data["name"], data["methods"])

    def __eq__(self, other):
        return self._name == other._name and self._methods == other._methods

    @property
    def description(self):
        return "Remove package %s" % self._name

    def needsclaims(self):
        yield "package:%s" % self._name

    def isneeded(self):
        # look for any of the facts saying we installed these things
        for method in _METHODS:
            localname = self._methods.get(method, self._name)
            factname = 'InstalledPackage:%s:%s' % (method, localname)
            if self._getfact(factname, False):
                return True
        return False

    def makechanges(self):
        # look for any of the facts saying we installed these things
        for method in _METHODS:
            localname = self._methods.get(method, self._name)
            factname = 'InstalledPackage:%s:%s' % (method, localname)
            if not self._getfact(factname, False):
                continue

            def defaultuninstall(name):
                return [method, 'uninstall', name]

            cmd = _UNINSTALL.get(method, defaultuninstall)(localname)
            if method in _ASROOT:
                if not allowinteractive():
                    raise HelperError("Need to be able to escalate to root")
                cmd.insert(0, 'sudo')
            try:
                execute(cmd)
            finally:
                # always clear the fact
                self._clearfact(factname)
        raise HelperError("Didn't remove package %s" % self._name)

    def wantspath(self, path):
        return False



import os

from homely._test import contents
from homely._test.system import HOMELY, TempRepo, checkrepolist, getsystemfn


def test_homely_remove(tmpdir, HOME):
    system = getsystemfn(HOME)

    def _addfake(name, createfile):
        # create a fake repo and add it
        tr = TempRepo(tmpdir, name)
        tf = os.path.join(HOME, createfile)
        contents(tr.remotepath + '/HOMELY.py',
                 """
                 from homely.files import lineinfile
                 lineinfile('~/%s', 'Hello from %s')
                 """ % (createfile, name))
        assert not os.path.exists(tf)
        system(HOMELY('add') + [tr.url])
        assert contents(tf) == "Hello from %s\n" % name
        return tr

    r1 = _addfake('repo1', 'file1.txt')
    r2 = _addfake('repo2', 'file2.txt')
    r3 = _addfake('repo3', 'file3.txt')

    # check that all the repos are there
    checkrepolist(HOME, system, [r1, r2, r3])
    assert contents(HOME + '/file1.txt', "Hello from repo1\n")
    assert contents(HOME + '/file2.txt', "Hello from repo2\n")
    assert contents(HOME + '/file3.txt', "Hello from repo3\n")

    # Check that the repo can be removed.
    system(HOMELY('forget') + [r1.repoid])
    checkrepolist(HOME, system, [r2, r3])
    assert contents(HOME + '/file1.txt', "Hello from repo1\n")
    assert contents(HOME + '/file2.txt', "Hello from repo2\n")
    assert contents(HOME + '/file3.txt', "Hello from repo3\n")

    # now run an update to make repo1's files go away
    system(HOMELY('update'))
    assert not os.path.exists(HOME + '/file1.txt')
    assert contents(HOME + '/file2.txt', "Hello from repo2\n")
    assert contents(HOME + '/file3.txt', "Hello from repo3\n")

    # Test removing multiple repos, but using local path this time
    # Note that because we don't use --update, the created files will still be
    # sitting around on disk
    system(HOMELY('forget') + ['~/repo2', '~/repo3'])
    checkrepolist(HOME, system, [])
    # repo2 and repo3 are stilling going to hang around on disk
    assert os.path.exists(HOME + '/repo2')
    assert os.path.exists(HOME + '/repo3')
    assert not os.path.exists(HOME + '/file1.txt')
    assert contents(HOME + '/file2.txt', "Hello from repo2\n")
    assert contents(HOME + '/file3.txt', "Hello from repo3\n")



import os

import pytest


def test_paths(tmpdir, HOME):
    from homely._utils import isnecessarypath
    l1 = os.path.join(tmpdir, 'link1')
    d1 = os.path.join(tmpdir, 'dir1')
    d1s = os.path.join(d1, 'subdir')
    l1s = os.path.join(l1, 'subdir')

    os.mkdir(d1)
    os.mkdir(d1s)
    os.symlink(d1, l1)
    assert os.path.isdir(l1s)

    assert isnecessarypath(tmpdir, d1)
    assert isnecessarypath(tmpdir, l1)
    assert isnecessarypath(l1, l1s)
    assert isnecessarypath(d1, l1s)
    assert not isnecessarypath(d1, l1)


fixed = [
    'http://www.foo.com/foo.txt',
    'git+ssh://git.example.com/example/bar',
    'git+ssh://foo@example.com/example/bar',
    'homely.test.repo:///foo/something/bar',
]


def test_expansion_homepath(tmpdir):
    from homely._utils import _homepath2real

    # test how various paths expand out to the home dir
    os.environ["HOME"] = tmpdir

    dir1 = os.path.join(tmpdir, 'dir1')
    link1 = tmpdir + "/link1"

    os.mkdir(dir1)
    os.symlink(dir1, link1)

    assert _homepath2real('.vimrc') == tmpdir + '/.vimrc'
    assert _homepath2real('.vimrc') == tmpdir + '/.vimrc'
    assert _homepath2real('bin/ack') == tmpdir + '/bin/ack'
    assert _homepath2real('link1/foo.txt') == tmpdir + '/link1/foo.txt'
    assert _homepath2real('/etc/hosts') == '/etc/hosts'

    assert _homepath2real('~/.vimrc') == tmpdir + '/.vimrc'
    assert _homepath2real('$HOME/.vimrc') == tmpdir + '/.vimrc'
    assert _homepath2real('~/bin/ack') == tmpdir + '/bin/ack'

    # check that relative paths like '.' also work
    assert _homepath2real('..') == '..'
    assert _homepath2real('../foo') == '../foo'

    # these should all raise errors
    badpaths = ['.', 'bin/', '/etc/']
    for path in badpaths:
        try:
            _homepath2real(path)
        except Exception:
            pass
        else:
            raise Exception("Expected an error from %r!" % path)

    for path in fixed:
        assert _homepath2real(path) == path


def test_expansion_repopath(tmpdir):
    from homely._utils import _repopath2real
    os.environ["HOME"] = tmpdir

    rdir = os.path.join(tmpdir, 'repo.git')

    # construct a fake repo
    from homely._vcs.testhandler import Repo
    repo = Repo(rdir, False, False, None)
    assert _repopath2real('foo.txt', repo) == rdir + '/foo.txt'
    assert _repopath2real('a/b/c.txt', repo) == rdir + '/a/b/c.txt'
    assert _repopath2real('/tmp/foo.txt', repo) == os.path.realpath('/tmp/foo.txt')
    # if '~' is included in the path then it will always expand out to $HOME
    assert _repopath2real('~/bin/foo.txt', repo) == tmpdir + '/bin/foo.txt'
    assert _repopath2real('$HOME/bin/foo.txt', repo) == tmpdir + '/bin/foo.txt'
    for path in fixed:
        assert _repopath2real(path, repo) == path


def test_time_interval_to_delta():
    from datetime import timedelta

    from homely._utils import _time_interval_to_delta

    # simple conversions
    assert _time_interval_to_delta('1h') == timedelta(hours=1)
    assert _time_interval_to_delta('27h') == timedelta(hours=27)
    assert _time_interval_to_delta('5d') == timedelta(days=5)
    assert _time_interval_to_delta('1125d') == timedelta(days=1125)
    assert _time_interval_to_delta('3w') == timedelta(weeks=3)
    assert _time_interval_to_delta('52w') == timedelta(weeks=52)

    # leading zeros are ignored
    assert _time_interval_to_delta('05d') == timedelta(days=5)

    # zero is technically valid
    assert _time_interval_to_delta('0d') == timedelta(seconds=0)

    # you can also pass a regular timedelta and get it back
    assert _time_interval_to_delta(timedelta(minutes=125)) == timedelta(minutes=125)

    def assert_invalid(invalid_value):
        with pytest.raises(ValueError, match="Invalid time interval"):
            _time_interval_to_delta(invalid_value)

    # negative values are invalid
    assert_invalid('-1h')
    # no number is not allowed
    assert_invalid('w')
    # no quantity is not allowed
    assert_invalid('5')
    # invalid type not allowed
    assert_invalid('5z')

    # ordinary integers are not allowed, but it raises a TypeError instead
    with pytest.raises(TypeError):
        _time_interval_to_delta(5)



import os

from homely._test import contents, waitfor
from homely._test.system import HOMELY, TempRepo, getjobstartfn, getsystemfn


def test_homely_updatestatus(HOME, tmpdir):
    from homely._utils import FAILFILE, RUNFILE, STATUSCODES, UpdateStatus

    system = getsystemfn(HOME)
    jobstart = getjobstartfn(HOME)

    exit_ok = STATUSCODES[UpdateStatus.OK]
    exit_never = STATUSCODES[UpdateStatus.NEVER]
    exit_paused = STATUSCODES[UpdateStatus.PAUSED]
    exit_failed = STATUSCODES[UpdateStatus.FAILED]
    exit_running = STATUSCODES[UpdateStatus.RUNNING]

    # create a special HOMELY.py script that we can control easily by putting magic files in the
    # right places
    spinfile = HOME + '/spin'
    diefile = HOME + '/diediedie'

    xdir1 = HOME + '/xdir1'
    assert not os.path.isdir(xdir1)

    # verify that homely updatestatus thinks we have never run an update
    system(HOMELY('updatestatus'), expecterror=exit_never)

    # pause autoupdate, and then check the status again
    system(HOMELY('autoupdate') + ['--pause'])
    system(HOMELY('updatestatus'), expecterror=exit_paused)

    # pausing again don't change anything
    system(HOMELY('autoupdate') + ['--pause'])
    system(HOMELY('updatestatus'), expecterror=exit_paused)

    # unpause takes us back to the previous status
    system(HOMELY('autoupdate') + ['--unpause'])
    system(HOMELY('updatestatus'), expecterror=exit_never)

    # add our special repo with our special file
    repo1 = TempRepo(tmpdir, 'repo1')
    contents(repo1.remotepath + '/HOMELY.py',
             """
             import os, time, sys
             from homely.files import mkdir
             while os.path.exists(%(spinfile)r):
                time.sleep(0.01)
             assert not os.path.exists(%(diefile)r), "Incredibly bad things in %(diefile)s"
             mkdir('~/xdir1')
             """ % locals())
    system(HOMELY('add') + [repo1.url])
    assert os.path.isdir(xdir1)

    # verify that homely updatestatus thinks we have run an update
    system(HOMELY('updatestatus'))

    # use the spinfile to make homely update sit in the background while we test some more things
    contents(spinfile, "spin!")
    with jobstart(HOMELY('update')) as job:
        for _ in waitfor('Appearance of RUNFILE %s' % RUNFILE):
            if os.path.exists(RUNFILE):
                break
        # verify that updatestatus says we are currently running
        system(HOMELY('updatestatus'), expecterror=exit_running)

        # remove the spinfile so the background job can finish
        os.unlink(spinfile)

    # verify that no update is currently running, and that the last run was successful
    system(HOMELY('updatestatus'))

    # now touch the errorfile and try a new update
    contents(diefile, "boom!")
    system(HOMELY('update'), expecterror=1)

    # updatestatus should tell us that the previous run failed
    system(HOMELY('updatestatus'), expecterror=exit_failed)

    # the error file should exist
    assert os.path.exists(FAILFILE)
    # grab the name of the OUTFILE
    outfile = system(HOMELY('autoupdate') + ['--outfile']).strip()
    # note that a plain 'homely update' won't create the outfile for us
    assert not os.path.exists(outfile)

    # try and kick off an autoupdate daemon ... it should fail miserably because the previous
    # update has already failed
    system(HOMELY('autoupdate') + ['--daemon'], expecterror=1)

    # use autoupdate --reset to allow automatic updates to resume
    system(HOMELY('autoupdate') + ['--clear'])
    system(HOMELY('updatestatus'), expecterror=exit_ok)

    # try and kick off an autoupdate daemon again ... it should fail miserably because the previous
    # update was too recent
    system(HOMELY('autoupdate') + ['--daemon'], expecterror=1)

    # remove the TIMEFILE so that homely thinks an update has never been run before
    from homely._utils import TIMEFILE
    os.unlink(TIMEFILE)

    try:
        # we use the spinfile to make sure the next autoupdate is going to stall
        contents(spinfile, "spin!")
        contents(diefile, "boom!")
        system(HOMELY('autoupdate') + ['--daemon'])

        # wait for the RUNFILE to appear
        for _ in waitfor('Appearance of RUNFILE'):
            if os.path.exists(RUNFILE):
                break

        # assert that our status is "running"
        system(HOMELY('updatestatus'), expecterror=exit_running)
    finally:
        # remove the spinfile so that the daemon can finish
        os.unlink(spinfile)

    # wait for the runfile to disappear
    for _ in waitfor('Disappearance of RUNFILE %s' % RUNFILE):
        if not os.path.exists(RUNFILE):
            break

    # the autoupdate should have failed because of the diefile
    system(HOMELY('updatestatus'), expecterror=exit_failed)
    with open(outfile) as f:
        assert 'Incredibly bad things' in f.read()

    # let's do this again, but without the diefile
    os.unlink(diefile)
    system(HOMELY('autoupdate') + ['--clear'])
    # we also need to manually remove the timefile
    os.unlink(TIMEFILE)
    try:
        contents(spinfile, "spin!")
        system(HOMELY('autoupdate') + ['--daemon'])
        for _ in waitfor('Appearance of RUNFILE'):
            if os.path.exists(RUNFILE):
                break
        system(HOMELY('updatestatus'), expecterror=exit_running)
    finally:
        os.unlink(spinfile)
    for _ in waitfor('Disappearance of RUNFILE %s' % RUNFILE):
        if not os.path.exists(RUNFILE):
            break

    # updatestatus should show that the update was successful
    system(HOMELY('updatestatus'), expecterror=exit_ok)



import functools
import os.path
import shutil
import sys
import tempfile
from typing import Optional

import pytest


@pytest.fixture(scope="function")
def HOME(tmpdir):
    old_home = os.environ['HOME']

    try:
        home = os.path.join(tmpdir, 'john')
        os.mkdir(home)
        # NOTE: homely._utils makes use of os.environ['HOME'], so we need to
        # destroy any homely modules that may have imported things based on this.
        # Essentially we blast away the entire module and reload it from scratch.
        for name in list(sys.modules.keys()):
            if name.startswith('homely.'):
                sys.modules.pop(name, None)
        os.environ['HOME'] = home
        yield home
    finally:
        os.environ['HOME'] = old_home


@pytest.fixture(scope="function")
def tmpdir(request):
    path = tempfile.mkdtemp()
    destructor = shutil.rmtree

    def destructor(path):
        print("rm -rf %s" % path)
        shutil.rmtree(path)
    request.addfinalizer(functools.partial(destructor, path))
    return os.path.realpath(path)


@pytest.fixture
def testrepo(HOME, tmpdir):
    # XXX: this fixture needs the "HOME" fixture or it won't be able to
    # homely-add the repo
    from homely._test.system import TempRepo
    yield _get_test_repo(TempRepo(tmpdir, 'cool-testrepo'))


def _get_test_repo(repo) -> None:
    from homely._ui import addfromremote
    from homely._utils import RepoListConfig, saveconfig
    from homely._vcs import Repo, testhandler

    handler: Optional[Repo] = testhandler.Repo.frompath(repo.url)
    assert handler is not None
    localrepo, needpull = addfromremote(handler, None)

    with saveconfig(RepoListConfig()) as cfg:
        cfg.add_repo(localrepo)

    return repo


@pytest.fixture
def testrepo2(HOME, tmpdir):
    # XXX: this fixture needs the "HOME" fixture or it won't be able to
    # homely-add the repo
    from homely._test.system import TempRepo
    yield _get_test_repo(TempRepo(tmpdir, 'cool-testrepo-2'))



import os
import re

import homely._vcs
from homely._errors import ConnectionError, RepoError, RepoHasNoCommitsError
from homely._utils import _expandpath, run
from homely.system import execute


class Repo(homely._vcs.Repo):
    type = homely._vcs.HANDLER_GIT_v1
    pulldesc = 'git pull'

    @classmethod
    def _from_parts(class_, repo_path, user, domain, name):
        if name.endswith('.git'):
            name = name[0:-4]

        canonical = 'https://%s/%s/%s.git' % (domain, user, name)

        return class_(
            repo_path,
            isremote=True,
            iscanonical=repo_path == canonical,
            canonical=canonical,
            suggestedlocal=name,
        )

    @classmethod
    def frompath(class_, repo_path):
        if os.path.isdir(repo_path):
            if not os.path.isdir(os.path.join(repo_path, '.git')):
                return
            return class_(_expandpath(repo_path),
                          isremote=False,
                          iscanonical=False,
                          suggestedlocal=None
                          )
        if (repo_path.startswith('ssh://') or
                repo_path.startswith('https://') or
                repo_path.startswith('git@')):

            # handle git@ url
            m = re.match(r'^git@([^/]+)[/:]([^/]+)/([^/]+)', repo_path)
            if m:
                domain, user, name = m.groups()
                return class_._from_parts(repo_path, user=user, domain=domain, name=name)

            # TODO: gitlab allows '/' in the org name so we actually want all but the last '/part'
            # to be the 'user'
            m = re.match(r'^https://([^/]+)/([^/]+)/([^/]+)', repo_path)
            if m:
                domain, user, name = m.groups()
                return class_._from_parts(repo_path, user=user, domain=domain, name=name)

            return class_(repo_path,
                          isremote=True,
                          iscanonical=False,
                          suggestedlocal=None)

    def pullchanges(self):
        assert not self.isremote
        cmd = ['git', 'pull']
        code, _, err = execute(cmd,
                               cwd=self.repo_path,
                               stderr=True,
                               expectexit=(0, 1))
        if code == 0:
            return

        assert code == 1
        needle = b'fatal: Could not read from remote repository.'
        if needle in err:
            raise ConnectionError()

        raise SystemError("Unexpected output from 'git pull': {}".format(err))

    def clonetopath(self, dest):
        origin = self.repo_path
        execute(['git', 'clone', origin, dest])

    def getrepoid(self):
        assert not self.isremote
        cmd = ['git', 'rev-list', '--max-parents=0', 'HEAD']
        returncode, stdout = run(cmd,
                                 cwd=self.repo_path,
                                 stdout=True,
                                 stderr="STDOUT")[:2]
        if returncode == 0:
            return self._getfirsthash(stdout)
        if returncode != 128:
            raise Exception("Unexpected returncode {} from git rev-list"
                            .format(returncode))

        if b"ambiguous argument 'HEAD'" not in stdout:
            raise Exception("Unexpected exitcode {}".format(returncode))

        # there's no HEAD revision, so we'll do the command again with
        # --all instead
        cmd = ['git', 'rev-list', '--max-parents=0', '--all']
        # use run() instead of execute() so that we don't print script output
        returncode, stdout = run(cmd,
                                 cwd=self.repo_path,
                                 stdout=True,
                                 stderr="STDOUT")[:2]
        if returncode == 0:
            if stdout == b'':
                raise RepoHasNoCommitsError()
            return self._getfirsthash(stdout)
        if returncode != 129:
            raise Exception("Unexpected returncode {} from git rev-list"
                            .format(returncode))
        if b"usage: git rev-list" in stdout:
            raise RepoHasNoCommitsError()

        raise SystemError("Unexpected exitcode {}".format(returncode))

    def _getfirsthash(self, stdout):
        stripped = stdout.rstrip().decode('utf-8')
        if '\n' in stripped:
            raise RepoError("Git repo has multiple initial commits")
        return stripped

    @staticmethod
    def shortid(repoid):
        return repoid[0:8]

    def isdirty(self):
        cmd = ['git', 'status', '--porcelain']
        out = execute(cmd, cwd=self.repo_path, stdout=True)[1]
        for line in out.split(b'\n'):
            if len(line) and not line.startswith(b'?? '):
                return True
        return False



import os

from homely._test import contents
from homely._test.system import HOMELY, TempRepo, checkrepolist, getsystemfn


def test_add_greenfield_repos(tmpdir, HOME):
    """
    Work or at least die with a helpful error message when ...
    """
    from homely._errors import ERR_NO_COMMITS, ERR_NO_SCRIPT, ERR_NOT_A_REPO

    system = getsystemfn(HOME)

    def _repolist():
        cmd = HOMELY('repolist') + ['--format', '%(localpath)s']
        return list(filter(None, system(cmd).strip().split("\n")))

    # Target dir is not a git repo
    notarepo = HOME + '/not_a_repo'
    os.mkdir(notarepo)
    contents(notarepo + '/HOMELY.py', "print('YES')\n")
    output = system(HOMELY('add') + [notarepo], expecterror=1)
    assert ERR_NOT_A_REPO in output
    del notarepo, output
    # this should have resulted in nothing being added to homely
    assert not len(_repolist())

    # Target dir doesn't have any commits
    nocommits = HOME + '/nocommits'
    os.mkdir(nocommits)
    contents(nocommits + '/HOMELY.py', "print('YES')\n")
    system(['git', 'init'], cwd=nocommits)
    assert os.path.exists(nocommits + '/.git')
    output = system(HOMELY('add') + [nocommits], expecterror=1)
    assert ERR_NO_COMMITS in output
    del nocommits, output
    # this should have resulted in nothing being added to homely
    assert not len(_repolist())

    # Target dir doesn't have a HOMELY.py script - when we add a repo that
    # doesn't have a HOMELY.py script we should raise an error and provide
    # instructions on how to write a HOMELY.py script (and not try to do the
    # homely update).
    noscript = HOME + '/noscript'
    os.mkdir(noscript)
    readme = noscript + '/README.md'
    contents(readme, "Hello world!\n")
    system(['git', 'init'], cwd=noscript)
    system(['git', 'config', 'user.name', "fred"], cwd=noscript)
    system(['git', 'config', 'user.email', "fred@example"], cwd=noscript)
    system(['git', 'add', 'README.md'], cwd=noscript)
    system(['git', 'commit', '-m', "Added readme"], cwd=noscript)
    output = system(HOMELY('add') + [noscript], expecterror=1)
    assert ERR_NO_SCRIPT in output
    # the repo should have added to homely
    assert [noscript] == _repolist()


def test_homely_add_repolist(tmpdir, HOME):
    system = getsystemfn(HOME)

    # make a fake repo - create a dir, a folder, and a symlink
    repo1 = TempRepo(tmpdir, 'repo1')
    homedir1 = os.path.join(HOME, 'dir1')
    homelink1 = os.path.join(HOME, 'link1')
    homedir1file = os.path.join(homelink1, 'file.txt')
    assert not os.path.exists(homedir1)
    assert not os.path.islink(homelink1)
    contents(repo1.remotepath + '/HOMELY.py',
             """
             from homely.files import lineinfile, mkdir, symlink
             mkdir('~/dir1')
             symlink('~/dir1', '~/link1')
             lineinfile('~/link1/file.txt', 'Hello World')
             """)

    # add the repo and ensure that everything was created as expected
    system(HOMELY('add') + [repo1.url])

    assert repo1.installedin(HOME)
    assert os.path.isdir(homedir1)
    assert os.path.islink(homelink1)
    assert os.readlink(homelink1) == os.path.realpath(homedir1)
    assert contents(homedir1file) == "Hello World\n"

    # make another repo that creates more things
    repo2 = TempRepo(tmpdir, 'repo2')
    repo2file = os.path.join(HOME, 'file2.txt')
    assert not os.path.exists(repo2file)
    contents(repo2.remotepath + '/HOMELY.py',
             """
             from homely.files import lineinfile
             lineinfile('~/file2.txt', 'Hey There')
             """)
    system(HOMELY('add') + [repo2.url])
    assert repo2.installedin(HOME)
    assert contents(repo2file, "Hey There\n")

    checkrepolist(HOME, system, [repo1, repo2])

    # a 3rd repo, but we're going to clone it into our home dir manually
    repo3 = TempRepo(tmpdir, 'repo3')
    contents(repo3.remotepath + '/HOMELY.py',
             """
             from homely.files import lineinfile
             lineinfile('~/r3.txt', 'From R3')
             """)
    # where would it go in the home dir?
    localrepo3 = os.path.join(HOME, 'repo3')
    # use a Repo instance to clone it into our home dir manually
    from homely._vcs.testhandler import Repo
    Repo.frompath(repo3.url).clonetopath(localrepo3)

    # test adding a repo from the local dir
    assert not os.path.exists(HOME + '/r3.txt')
    system(HOMELY('add') + [localrepo3])
    assert contents(HOME + '/r3.txt') == 'From R3\n'
    checkrepolist(HOME, system, [repo1, repo2, repo3])

    # test that you can't add the same repo again
    system(HOMELY('add') + [repo2.url])
    checkrepolist(HOME, system, [repo1, repo2, repo3])

    # test that adding a repo with something like 'homely add .' doesn't record
    # a stupid path like '.'
    repo4 = TempRepo(tmpdir, 'repo4')
    contents(repo4.remotepath + '/HOMELY.py',
             """
             from homely.files import lineinfile
             lineinfile('~/r4.txt', 'From R4')
             """)
    localrepo4 = os.path.join(HOME, 'repo4')
    # use a Repo instance to clone it into our home dir manually
    from homely._vcs.testhandler import Repo
    Repo.frompath(repo4.url).clonetopath(localrepo4)
    system(HOMELY('add') + ['.'], cwd=localrepo4)
    checkrepolist(HOME, system, [repo1, repo2, repo3, repo4])



import os.path

from homely._test import contents, run_update_all


def test_files_writefile(HOME, testrepo):
    contents(
        testrepo.remotepath + '/HOMELY.py',
        """
        import json
        from homely.files import mkdir, writefile

        mkdir('~/somedir')
        with writefile('~/somedir/somefile.txt') as f:
            json.dump({'one': 1*1}, f)
        """
    )

    # pull first so we get the new HOMELY.py script
    run_update_all(pullfirst=True)
    assert contents(HOME + '/somedir/somefile.txt') == '{"one": 1}'


def test_files_writefile_cleans_up_or_not(testrepo, HOME):
    # create file A.txt already
    contents(HOME + '/A.txt', 'before-update')

    contents(
        testrepo.remotepath + '/HOMELY.py',
        """
        from homely.files import writefile

        with writefile('~/A.txt') as f1:
            f1.write('AAA\\n')
        with writefile('~/B.txt') as f2:
            f2.write('BBB\\n')
        with writefile('~/C.txt') as f3:
            f3.write('CCC\\n')
        """
    )

    # pull first so we get the new HOMELY.py script
    run_update_all(pullfirst=True, cancleanup=True)
    assert contents(HOME + '/A.txt') == 'AAA\n'
    assert contents(HOME + '/B.txt') == 'BBB\n'
    assert contents(HOME + '/C.txt') == 'CCC\n'

    # now remove the writes to A.txt and B.txt
    contents(
        testrepo.remotepath + '/HOMELY.py',
        """
        from homely.files import writefile

        with writefile('~/C.txt') as f3:
            f3.write('CCC FINAL\\n')
        """
    )
    run_update_all(pullfirst=True, cancleanup=True)

    # A.txt is *not* cleaned up because it wasn't created by homely
    assert contents(HOME + '/A.txt') == 'AAA\n'

    # B.txt *is* cleaned up because homely creatd it
    assert not os.path.exists(HOME + '/B.txt')

    # C.txt is modified to final state
    assert contents(HOME + '/C.txt') == 'CCC FINAL\n'



#!/usr/bin/env python3
import glob
import os
import shutil
import sys
from datetime import date
from subprocess import check_call, check_output


def fail(message):
    sys.stderr.write(message)
    sys.stderr.write("\n")
    sys.exit(1)


def updatechangelog(new_tag, latest_tag):
    # what should the heading at top of changelog be?
    new_heading = "Version {} - {}".format(
        new_tag, date.today().strftime("%d %b %Y"))

    changelog_old = 'CHANGELOG.rst'
    changelog_new = changelog_old + '.new'

    heading_replaced = False
    replace_next_lines = False

    with open(changelog_old) as f_old, open(changelog_new, 'w') as f_new:
        for line in f_old:
            if line == 'NEW\n':
                # instead of copying this line, replace it with the new heading
                f_new.write(new_heading)
                f_new.write('\n')
                heading_replaced = True
                replace_next_lines = True
                continue

            if replace_next_lines:
                replace_next_lines = False
                f_new.write('-' * len(new_heading))
                f_new.write('\n')

                if line == '\n':
                    # a blank line should be re-inserted
                    f_new.write('\n')
                if not line.startswith('---'):
                    # Don't re-insert the dashed line (it will have incorrect
                    # length). Anything else is unexpected.
                    fail('"NEW" heading not followed by blank line or "---"')

                continue

            # otherwise, just copy the line to the new file
            f_new.write(line)

    # if we didn't replace the NEW heading, we should add it
    if not heading_replaced:
        with open(changelog_old) as f_old, open(changelog_new, 'w') as f_new:
            heading_added = False
            seen_title = False
            for line in f_old:
                f_new.write(line)
                if heading_added:
                    pass
                elif line.strip() == 'CHANGELOG':
                    seen_title = True
                elif seen_title and line.startswith('==='):
                    # add the new heading here
                    f_new.write('\n\n')
                    f_new.write(new_heading)
                    f_new.write('\n')
                    f_new.write('-' * len(new_heading))
                    f_new.write('\n\n')
                    f_new.write('* DOCUMENT CHANGES HERE\n')
                    heading_added = True

        if not heading_added:
            fail('"NEW" heading not present and was not added automatically')

    # have a go at editing the file directly
    splitcmds = [
        # start a new buffer
        '+new',
        # give it a name
        '+silent file === Changes since {} ==='.format(latest_tag),
        # paste git log output
        '+silent read !git log --oneline {}..'.format(latest_tag),
        # delete blank line at top
        '+normal! ggdd',
        # modify buffer settings so that vim doesn't ask to save this buffer
        '+setlocal buftype=nofile',
        # jump back to original window
        '+wincmd w',
    ]
    check_call(['vim', changelog_new] + splitcmds)

    # move file sideways and commit it
    shutil.move(changelog_new, changelog_old)
    check_call(['git', 'commit', changelog_old, '-m', 'Update changelog'])


def main():
    tag = sys.argv[1]
    if not len(tag):
        fail("USAGE: {} TAG".format(sys.argv[0]))

    # get the latest tag name
    revlist = check_output(['git', 'rev-list', '--tags', '--max-count=1'])
    branches = filter(None, revlist.decode('utf-8').split('\n'))
    cmd = ['git', 'describe', '--tags'] + list(branches)
    latest_tag = check_output(cmd).strip().decode('utf-8')

    # make sure there is nothing in git status that could get in the way
    output = check_output(['git', 'st', '-s'])
    if len(output):
        fail("ERROR: A clean checkout is required")

    updatechangelog(tag, latest_tag)

    # create the new tag
    check_call(['git', 'tag', tag, '-m', 'Tagged {}'.format(tag)])
    # build
    check_call(['python3', 'setup.py', 'sdist'])
    check_call(['python3', 'setup.py', 'bdist_wheel', '--universal'])
    # upload
    files = ['dist/homely-{}.tar.gz'.format(tag)]
    files.extend(glob.glob('dist/homely-{}-py*.whl'.format(tag)))

    repository = 'https://upload.pypi.org/legacy/'

    check_call(['twine', 'upload', '--repository-url', repository] + files)
    print("NEW VERSION PUBLISHED!")
    check_call(['git', 'push'])
    check_call(['git', 'push', '--tags'])


if __name__ == '__main__':
    main()
